#ifndef _ENTITY_H_
#define _ENTITY_H_
/*******************************************************************************
  Copyright (c) 1998-2007 J. P. Abelanet/Basis Software.
  All Rights Reserved.

  -> THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF BASIS SOFTWARE. <-

  FILE
    Entity.h
    
  DESCRIPTION
    This is the abstract base class for persistent objects in a modeling session.
    
  AUTHOR
    J. P. Abelanet - 04/18/98
*******************************************************************************/

/* system includes */
#include <stdio.h>

/* library includes */
#include <geom/ClassID.h>
#include <vmath/Array.h>
#include <vmath/Base.h>
#include <vmath/Error.h>
#include <vmath/ErrorLogger.h>


/*******************************************************************************
  PRIVATE FUNCTIONS - internal use - not intended to be called directly
*******************************************************************************/
//#define ENT_VERBOSE_METHODS
/*******************************************************************************
  PUBLIC FUNCTIONS - can be called freely by anyone
*******************************************************************************/



/*******************************************************************************
  PRIVATE FUNCTIONS - internal use - not intended to be called directly
*******************************************************************************/
/* forward declarations for ::friends */
class AttributeChangeNotifier;
class EntityCopier;
class EntityTransformer;
class NativeTopologyReaderPre5;
/*******************************************************************************
  PUBLIC FUNCTIONS - can be called freely by anyone
*******************************************************************************/


/*******************************************************************************
  ABSTRACT
    This is the namespace for the GE toolkit.

  AUTHOR
    J. P. Abelanet - 04/07/05
*******************************************************************************/
namespace GE {

/* forward declarations */
template <class T> class EntityList;
class Attribute;
class AttributeRegistry;
class BSplineCurve;
class Database;
class EntitySupportScanner;
class UIDMapper;
class ForeignEntityMapper;
class IntervalVector3D;
class NativeFileWriter;
class Region;
class TransformMatrix;
class UnsharedAttribute;




/*******************************************************************************
  ABSTRACT
    This is the abstract base class for persistent objects in a modeling session.
    The address of an entity will be the same for the life of the entity.
    If two pointers to entities are equal, they reference the exact same
    entity.

  NOTES
    - Subclasses must have a call to the macro GE_ENTITY_CLASS_HEADER()
      placed inside the class definition, and must have a call to the macro
      GE_ENTITY_CLASS_BODY() placed inside the implementation file of the class
    - Every element that needs to be written to file that can be linked in a
      non-linear list must be a subclass of this.
    - This includes (but is not limited to):
      - elements that can be pointed to by multiple relatives
      - any element that has more than one of: parents, siblings, children
        (ie - a parent and a sibling)

  REVISIONS
    04/28/01 - J. P. - Changed from void* to void
    10/11/04 - J. P. - Made base class for all entities and added base data
                       to this class
    03/16/05 - J. P. - Removed unused Magic_ field and added DeletePending_
    11/30/06 - J. P. - Modified to use ClassID class, removed classID field,
                       and resize marker_
    03/26/07 - J. P. - Made marker_ UInt32 and made all marker access via
                       EntitySet
    
  AUTHOR
    J. P. Abelanet - 04/18/98
*******************************************************************************/
class Entity {

public:	// public classes

/*******************************************************************************
  ABSTRACT
    Class containing errors that were found in an Entity during integrity checking.
    
  BUGS
    This should be a class with the following fields:
    - entity (Entity that contained the error)
    - entity2 (Entity that causes a problem for the first Entity - usually NULL)
    - error (Error value of this error)
    - coordinate (3-D location of the error)

  AUTHOR
    J. P. Abelanet - 12/05/05
*******************************************************************************/
  typedef Error EntityError;
  
public:	// public typedefs

/*******************************************************************************
  ABSTRACT
    This is the predefined type for Universal IDentifiers for Entities.
    This is primarily used during native file export and import in order to
    allow foreign entities to maintain a persistent link to a native Entity.

  NOTES
    - This is guaranteed to have the same size as sizeof(void*) to allow this
      to be a temporary alias for a pointer.
    
  REVISIONS
    01/04/07 - J. P. - Modified to work for 64-bit platforms
    
  AUTHOR
    J. P. Abelanet - 04/18/98
*******************************************************************************/
  typedef UIntPtr	UID;

  static const UID	UID_NULL = 0;	/* NULL entity UID      */

public:	// public accessor methods
  Database *Inquire_Database () const;
  Error Detach_Database () const;

  Boolean Inquire_Deletable (Entity **const DependentChild) const;

  Attribute *Inquire_FirstAttribute (const AttributeRegistry *attributeRegistry) const;
  void Inquire_Attributes (const AttributeRegistry *attributeRegistry,
           EntityList<Attribute> *const attributes,
	   Int32 *const numAttributes = NULL) const;
  void Print_Attributes (const AttributeRegistry *attributeRegistry,
           FILE *strm) const;

public:	// public mutator methods
  Error Delete ();

public:	// public static methods
  static Error Copy_Multiple (const EntityList<Entity> *entityList,
         EntityList<Entity> *const EntityCopies, const Boolean shareSupportEntities = FALSE);

  static Error Apply_Transform_Multiple (EntityList<Entity> *entityList,
         const TransformMatrix& Matrix);

  static Error Delete_Multiple (EntityList<Entity> *const entityList);

public:	// public virutal methods
  virtual Boolean Is_Major () const;

  virtual Entity *Copy (const Boolean shareSupportEntities = FALSE) const;

  virtual Error Apply_Transform (const TransformMatrix& Matrix);

public:	// public pure virtual methods
/*******************************************************************************
  ABSTRACT
    This routine describes an Entity.
  
  PARAMETERS
    returns:
      Ptr to description - internal static data that cannot be modified

  REVISIONS
    10/02/02 - J. P. - Modified to call Entity::Inquire_Description
    11/12/02 - J. P. - Added ST/F/E/V
    02/19/03 - J. P. - Added distinction between group and instance
    02/26/03 - J. P. - Added support for attribute templates
    03/10/03 - J. P. - Added explicit casts
    03/21/03 - J. P. - Added coedge
    05/06/04 - J. P. - Adding explicit casts in preparation for conversion to
                       C++ classes
    11/15/04 - J. P. - Modified to use classes
    11/19/04 - J. P. - Modified to return const char *
    
  AUTHOR
    J. P. Abelanet - 08/06/01
*******************************************************************************/
  virtual const char *Inquire_Description () const = 0;

/*******************************************************************************
  ABSTRACT
    This routine finds the 3-D axis aligned bounding box that encloses the
    specified Entity.
  
  PARAMETERS
    ExtentBox  OUT
      3-D bounding box
  
    returns:
      Error_invalid_entity	- entity type not valid
      Error_no_extents		- compound entity contains no entities whose extents can be computed

  REVISIONS
    03/10/03 - J. P. - Added explicit casts
    01/27/04 - J. P. - Modified to call Model::Inquire_ExtentBox
    05/06/04 - J. P. - Adding explicit casts in preparation for conversion to
                       C++ classes
    11/12/04 - J. P. - Modified to use classes
    08/18/05 - J. P. - Modified to use Interval class
    
  AUTHOR
    J. P. Abelanet - 10/02/02
*******************************************************************************/
  virtual Error Inquire_ExtentBox (IntervalVector3D& ExtentBox) const = 0;

/*******************************************************************************
  ABSTRACT
    This method checks the integrity of an Entity.
  
  PARAMETERS
    options  IN
      Optional parameter that defaults to 0 -
      if non-0, bitwise-OR'ed combination of options - NOT USED currently
      
    entityErrors  OUT
      Optional parameter that defaults to NULL -
      if non-NULL, this will be the list of errors located in Entity
      
    returns:
      Error_no_errors	- no errors occurred

  NOTES
    It is currently possible for this to return a non-Error_no_errors value,
    indicating a problem with the Entity, but also return an empty list of
    entityErrors.
  
  REVISIONS
    11/16/06 - J. P. - Modified to use new Array template class

  AUTHOR
    J. P. Abelanet - 12/05/05
*******************************************************************************/
  virtual Error Check_Integrity (const Int32 options = 0,
          Array<EntityError> *const entityErrors = NULL) const = 0;

public:	// public virtual methods

/*******************************************************************************
  ABSTRACT
    This routine returns the ClassID for the Entity.
  
  PARAMETERS
    returns:
      ClassID of the given Entity
  
  REVISIONS
    11/30/06 - J. P. - Modified to use ClassID class
    
  AUTHOR
    J. P. Abelanet - 10/12/04
*******************************************************************************/
  virtual const ClassID& Inquire_ClassID () const = 0;

  virtual Boolean Is_InClassID (const ClassID& classID) const;

public:	// public static methods implemented on all classes
  static const ClassID& Static_ClassID ();
  static const Entity* Downcast (const Entity* entity);
  static Entity* Downcast (Entity* entity);

protected:	// protected methods
  inline Entity ();
  inline Entity (const Entity &Original);
  inline Entity& operator= (const Entity &old_class);
  inline virtual ~Entity ();

/*******************************************************************************
  PRIVATE FUNCTIONS - internal use - not intended to be called directly
*******************************************************************************/
public:	// private methods for EntitySet class
  Boolean Inquire_Marked_ (UInt32 marker) const		{ return ((marker_ & marker) != 0); }
  void Set_Marker_ (UInt32 marker) const		{ marker_ |= marker; }
  void Unset_Marker_ (UInt32 marker) const		{ marker_ &= ~marker; }
/*******************************************************************************
  PUBLIC FUNCTIONS - can be called freely by anyone
*******************************************************************************/

private:	// private enums
  enum { DELETEPENDING_MARKER = 0x80000000 };
  
private:	// private methods
  Boolean Inquire_DeletePending () const		{ return Inquire_Marked_ (DELETEPENDING_MARKER); }
  void Set_DeletePending () const			{ Set_Marker_ (DELETEPENDING_MARKER); }
  
  UnsharedAttribute *Inquire_Attribute () const		{ return (Att_); }
  void Assign_Attribute (UnsharedAttribute *Attribute)	{ Att_ = Attribute; }
  
private:	// private virtual methods - most are pure virtual
  virtual Boolean Inquire_Bound () const;
  
/*******************************************************************************
  ABSTRACT
    This routine destroys an Entity, assuming that all references to this Entity
    have been removed.
  
  PARAMETERS
    returns:
      Error_invalid_entity	- entity type not valid
      Error_invalid_entity	- entity not in database
      Error_still_referenced	- entity still referenced by another entity
  
  NOTES
    - This should rarely need to be implemented, since this should be a thin
      wrapper for the destructor
    - The reason for using this versus using Entity::operator delete is that
      an error code is returned here.
    
  REVISIONS
    03/23/01 - J. P. - Modified to use backpointers
    04/28/01 - J. P. - Changed Entity from void* to void
    07/28/01 - J. P. - Modified to detach entity from groups and to check
                       for remaining associativity
    08/02/01 - J. P. - Added support for circles, revolutions, and extrusions
    10/02/02 - J. P. - Modified to call Entity::Delete
    02/26/03 - J. P. - Modified to handle attributes and attribute templates
    02/28/03 - J. P. - Modified to pass Entity::Static_ClassID () to ChildEntity::Entity_Inquire_Children
    03/10/03 - J. P. - Added explicit casts
    03/26/03 - J. P. - Swapped parameters to Group::Attach/Detach_Entity
    05/06/04 - J. P. - Adding explicit casts in preparation for conversion to
                       C++ classes
    11/11/04 - J. P. - Modified to use classes
    
  AUTHOR
    J. P. Abelanet - 02/11/01
*******************************************************************************/
  virtual Error Internal_Delete ()	{ delete this; return Error_no_errors; }
  
/*******************************************************************************
  ABSTRACT
    This routine transforms an Entity, without transforming any of the supporting
    Entities.  Supporting Entities are managed by an EntityTransformer.
  
  PARAMETERS
    Matrix  IN
      Transformation matrix to apply
  
    returns:
      Error_invalid_entity	- entity type not valid

  AUTHOR
    J. P. Abelanet - 05/01/06
*******************************************************************************/
  virtual Error Internal_Transform (const TransformMatrix& Matrix) = 0;

/*******************************************************************************
  ABSTRACT
    This routine copies an Entity, without copying any of the supporting
    Entities.  This should typically just invoke the copy constructor, creating
    a shallow copy of the Entity, since references to supporting Entities are
    managed by an EntityCopier.

  PARAMETERS
    returns:
      Copy of Entity -
      will be NULL if this routine fails

  NOTES
    - This should be a thin wrapper for the copy constructor
    - The reason for using this versus using the copy constructor is that
      this is a virutal function, and each implementation returns a pointer
      of the same type as "this"
    
  REVISIONS
    05/06/04 - J. P. - Adding explicit casts in preparation for conversion to
                       C++ classes
    11/24/04 - J. P. - Modified to use classes and to return the copy
    04/28/06 - J. P. - Modified to always use thin wrapper for copy constructor

  AUTHOR
    J. P. Abelanet - 10/02/02
*******************************************************************************/
  virtual Entity *Internal_Copy () const = 0;

/*******************************************************************************
  ABSTRACT
    This routine scans all support Entities referenced to by an Entity and
    adds them to an EntitySupportScanner if not already added.
    This is done by calling EntitySupportScanner::Scan_UnshareableSupportEntity()
    for each support Entity that is never shared, and by calling
    EntitySupportScanner::Scan_ShareableSupportEntity() for each support Entity that
    can be shared.
      
  PARAMETERS
    entitySupportScanner  INOUT
      EntitySupportScanner containing scanned Entities
  
    returns:
      Error_no_errors	- no errors occurred
  
  REVISIONS
    11/24/04 - J. P. - Modified to use classes
    12/14/04 - J. P. - Modified to return Error
    04/28/06 - J. P. - Modified to pass around an UIDMapper
    04/23/07 - J. P. - Modified to use EntitySupportScanner class
    
  AUTHOR
    J. P. Abelanet - 07/07/00
*******************************************************************************/
  virtual Error Scan_SupportEntities (EntitySupportScanner *const entitySupportScanner) const = 0;

/*******************************************************************************
  ABSTRACT
    This method converts all UIDs stored as pointers in an Entity
    to the equivalent pointers to Entities, using UIDMapper::Map_Entity().
    This is used during native file import and during Entity copying.
      
  PARAMETERS
    uidMapper  IN
      UIDMapper containing UID maps
  
    returns:
      Error_no_errors	- no errors occurred
  
  NOTES
    - If subclasses of ChildEntity are used
      to maintain backpointers to this Entity, this is where they should be
      created.
  
  REVISIONS
    11/24/04 - J. P. - Modified to use classes
    12/14/04 - J. P. - Modified to return Error
    02/07/06 - J. P. - Modified to take UIDMapper
    04/28/06 - J. P. - Made UIDMapper const
    
  AUTHOR
    J. P. Abelanet - 01/28/97
*******************************************************************************/
  virtual Error Map_Entities (const UIDMapper *uidMapper) = 0;

/*******************************************************************************
  ABSTRACT
    This is the function that is called on a native Entity after
    Entity::Map_Entities() has been called on all Entities.  This is where
    information that is not committed to file can be fully restored, but not
    without being able to access other Entities.  After this call, all the
    pointers on an Entity should be valid.
  
  PARAMETERS
    returns:
      Error_no_errors	- no errors occurred
  
  NOTES
    - It is rarely necessary to implement this method - the only core Entity
      that implements this is Model
    - Entities cannot be created or deleted in here -
      use NativeFileReader::DestructivelyPostProcessEntities() for that!
    
  AUTHOR
    J. P. Abelanet - 10/15/04
*******************************************************************************/
  virtual Error Update_PostNativeRead ();
  
/*******************************************************************************
  ABSTRACT
    This routine writes an entity to a native file.
      
  PARAMETERS
    nativeFileWriter  INOUT
      Native file to write to

    returns:
      Error_formatting_error	- unexpected error occured while formatting the native file
      Error_insufficient_memory	- unable to allocate memory
  
  NOTES
    This should always call NativeFileWriter::WriteEntityHeader() before writing the
    entity-specific data.
  
  AUTHOR
    J. P. Abelanet - 12/02/04
*******************************************************************************/
  virtual Error Write_Native (NativeFileWriter *const nativeFileWriter) const = 0;

private:	// private data members
  mutable UInt32	marker_;	// Entity marker used to indicate containment in EntitySets -
					// 0th bit used to indicate DeletePending -
					// - TRUE if entity being deleted or merged
  					// - Used to prevent premature deletion in BindingAttribute code
  UnsharedAttribute	*Att_;		// singly-linked list of Attributes for this Entity

/*******************************************************************************
  PRIVATE FUNCTIONS - internal use - not intended to be called directly
*******************************************************************************/
friend class BSplineCurve;		// for BSplineCurve::Make_ParameterSpaceBSpline
friend class Database;
friend class EntitySupportScanner;	// for Scan_SupportEntities()
friend class UIDMapper;			// for Map_Entities() and Inquire_Bound()
friend class BindingAttribute;		// for Inquire_Bound()
friend class ForeignEntityMapper;	// for Inquire_Bound()
friend class Model;			// for Inquire_Bound()
friend class NativeFileReader;
friend class NativeFileWriter;
friend class SharedAttribute;
friend class UnsharedAttribute;

friend class Coedge;
friend class Edge;
friend class Face;
friend class Loop;
friend class GE::Region;	// GE:: qualifier to avoid collisions with /usr/include/X11/Xutil.h/Region typedef
friend class Sheet;
friend class Shell;
friend class Vertex;

friend class ::AttributeChangeNotifier;	// for Set_DeletePending()
friend class ::EntityCopier;
friend class ::EntityTransformer;
friend class ::NativeTopologyReaderPre5;
/*******************************************************************************
  PUBLIC FUNCTIONS - can be called freely by anyone
*******************************************************************************/
};


/*******************************************************************************
  ABSTRACT
    This macro should be placed inside the class definition of all subclasses
    of Entity.  It defines all the methods related to ClassID's and safe casting.
  
  PARAMETERS
    theClass  IN
      Name of the class
  
  AUTHOR
    J. P. Abelanet - 11/30/06
*******************************************************************************/
#define GE_ENTITY_CLASS_HEADER(theClass)				\
									\
public:									\
  virtual const GE::ClassID& Inquire_ClassID () const;			\
  virtual GE::Boolean Is_InClassID (const GE::ClassID& classID) const;	\
  static const GE::ClassID& Static_ClassID ();				\
  static const theClass* Downcast (const GE::Entity* entity);		\
  static theClass* Downcast (GE::Entity* entity);


/*******************************************************************************
  ABSTRACT
    This macro should be placed inside the implementation file of all subclasses
    of Entity.  It implements all the methods related to ClassID's and safe casting.
  
  PARAMETERS
    theClass  IN
      Name of the class
  
    parentClass  IN
      Name of the parent class
  
    theName  IN
      Name of the class expressed as a C string - should be the same as theClass
      enclosed in double quotes
      
  AUTHOR
    J. P. Abelanet - 11/30/06
*******************************************************************************/
#define GE_ENTITY_CLASS_BODY(theClass, parentClass, theName)		\
									\
const GE::ClassID& theClass::Inquire_ClassID () const			\
{                                                                       \
  return theClass::Static_ClassID();					\
}                                                                       \
                                                                        \
GE::Boolean theClass::Is_InClassID (const GE::ClassID& classID) const	\
{                                                                       \
  if (classID == theClass::Static_ClassID ())				\
    return TRUE;							\
  if (classID.Inquire_ParentClassID () == parentClass::Static_ClassID ())	\
    return FALSE;							\
  return parentClass::Is_InClassID (classID);				\
}                                                                       \
                                                                        \
const GE::ClassID& theClass::Static_ClassID ()				\
{                                                                       \
  static const GE::ClassID classID_ (theName,				\
  			parentClass::Static_ClassID ());		\
									\
  return classID_;							\
}                                                                       \
                                                                        \
const theClass* theClass::Downcast (const GE::Entity* entity)		\
{                                                                       \
  if (entity && entity->Is_InClassID (theClass::Static_ClassID ()))	\
    return (const theClass*)entity;					\
  return NULL;								\
}                                                                       \
                                                                        \
theClass* theClass::Downcast (GE::Entity* entity)			\
{                                                                       \
  if (entity && entity->Is_InClassID (theClass::Static_ClassID ()))	\
    return (theClass *)entity;						\
  return NULL;								\
}


inline Entity::Entity ():
	marker_ (0),
	Att_ (NULL)
/*******************************************************************************
  ABSTRACT
    Default constructor for an Entity.
  
  PARAMETERS
    none
  
  REVISIONS
    03/16/05 - J. P. - Added DeletePending_ field
    03/26/07 - J. P. - Incorporated DeletePending_ field into marker_
    
  AUTHOR
    J. P. Abelanet - 10/15/04
*******************************************************************************/
{
#if defined(ENT_VERBOSE_METHODS)
  fprintf (ErrorLogger::GetNoteFile (), "Entering Entity::Entity\n");
#endif
} /* Entity::Entity */


inline Entity::Entity (const Entity &/*Original*/):
	marker_ (0),
	Att_ (NULL)
/*******************************************************************************
  ABSTRACT
    Default copy constructor for an Entity.
  
  PARAMETERS
    Original  IN
      Entity to be copied
  
  REVISIONS
    05/02/06 - J. P. - Modified to copy ClassID_
    11/30/06 - J. P. - ClassID no longer stored on each Entity
    03/26/07 - J. P. - Incorporated DeletePending_ field into marker_
    
  AUTHOR
    J. P. Abelanet - 10/26/04
*******************************************************************************/
{
#if defined(ENT_VERBOSE_METHODS)
  fprintf (ErrorLogger::GetNoteFile (), "Entering Entity::Entity (const Entity &Original)\n");
#endif
} /* Entity::Entity */


inline Entity& Entity::operator= (const Entity &old_class)
/*******************************************************************************
  ABSTRACT
    Assignment operator for an Entity.
  
  PARAMETERS
    old_class  IN
      Entity to be copied
  
  AUTHOR
    J. P. Abelanet - 03/23/05
*******************************************************************************/
{
#if defined(ENT_VERBOSE_METHODS)
  fprintf (ErrorLogger::GetNoteFile (), "Entering Entity::operator=\n");
#endif
  // Check for self-assignment
  if (this != &old_class) {
  
    // Is this legal?
    if (Inquire_ClassID () != old_class.Inquire_ClassID ())
      fprintf (ErrorLogger::GetErrorFile (), "ERROR:  Entity::operator=:  Cannot use between different classes\n");
  
    /* Copy the data from the source to the destination */
  }
  
  return (*this);
} /* Entity::operator= */


inline Entity::~Entity ()
/*******************************************************************************
  ABSTRACT
    Default destructor for an Entity.
  
  PARAMETERS
    none
  
  REVISIONS
    01/07/05 - J. P. - Added more descriptive feedback on error
    02/02/06 - J. P. - Modified to call Entity::Inquire_Description()
    11/30/06 - J. P. - Removed access to ClassID
    
  AUTHOR
    J. P. Abelanet - 10/15/04
*******************************************************************************/
{
#if defined(ENT_VERBOSE_METHODS)
  fprintf (ErrorLogger::GetNoteFile (), "Entering Entity::~Entity\n");
#endif
  /* Check for remaining attributes - this is an error */
  if (Inquire_Attribute ())
    fprintf (ErrorLogger::GetErrorFile (), "INTERNAL ERROR:  Entity::~Entity:  Attribute (%s) still attached to Entity (%p)!\n",
        ((const Entity*)Inquire_Attribute ())->Inquire_Description (),
	this);
} /* Entity::~Entity */


inline Error Entity::Update_PostNativeRead ()
{
#if defined(ENT_VERBOSE_METHODS)
  fprintf (ErrorLogger::GetNoteFile (), "Entering Entity::Update_PostNativeRead\n");
#endif
  return (Error_no_errors);
} /* Entity::Update_PostNativeRead */


inline Error Entity::Check_Integrity (const Int32 /*options*/,
       Array<EntityError> *const entityErrors) const
{
  // Init the outgoing parameters
  if (entityErrors)
    entityErrors->Clear ();
  return Error_no_errors;
} // Entity::Check_Integrity

}	// namespace GE

#endif /* !_ENTITY_H_ */
