/*******************************************************************************
  Copyright (c) 1997-2007 J. P. Abelanet/Basis Software.
  All Rights Reserved.

  -> THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF BASIS SOFTWARE. <-

  FILE
    Entity.cpp
    
  DESCRIPTION
    This is the abstract base class for persistent objects in a modeling session.
  
  AUTHOR
    J. P. Abelanet - 04/29/01
*******************************************************************************/

/* system includes */
#include <stdio.h>

/* library includes */
#include <geom/Attribute.h>
#include <geom/AttributeRegistry.h>
#include <geom/ClassID.h>
#include <geom/Database.h>
#include <geom/Entity.h>
#include <geom/EntityList.h>
#include <geom/EntityReferenceList.h>
#include <geom/StronglyBindingChild.h>
#include <geom/UnsharedAttribute.h>
#include <vmath/Base.h>
#include <vmath/Error.h>
#include <vmath/ErrorLogger.h>
#include <vmath/TransformMatrix.h>
#include <../src/geom/AttributeChangeNotifier.h>
#include <../src/geom/EntityCopier.h>
#include <../src/geom/EntityTransformer.h>

/* local declarations */



namespace GE {


Boolean Entity::Is_InClassID (const ClassID& classID) const
/*******************************************************************************
  ABSTRACT
    This routine determines if this Entity is in the given ClassID
    or one of its subclasses.
  
  PARAMETERS
    classID  IN
      ClassID to use
      
    returns:
      TRUE if this Entity is in the given ClassID or one of its subclasses
      
  REVISIONS
    11/30/06 - J. P. - Modified to use ClassID class
    
  AUTHOR
    J. P. Abelanet - 01/18/06
*******************************************************************************/
{
  return (classID == Entity::Static_ClassID ());
} // Entity::Is_InClassID


const ClassID& Entity::Static_ClassID ()
/*******************************************************************************
  ABSTRACT
    This returns the unique ClassID associated with subclasses of Entity.

  PARAMETERS
    returns:
      Unique ClassID associated with this subclass
      
  AUTHOR
    J. P. Abelanet - 11/30/06
*******************************************************************************/
{
  static const ClassID classID_ ("Entity");

  return classID_;
} // Entity::Static_ClassID


const Entity* Entity::Downcast (const Entity* entity)
/*******************************************************************************
  ABSTRACT
    This performs safe downcasting of subclasses of Entity.

  PARAMETERS
    entity  IN
      const Entity to use
    
    returns:
      Entity cast to this subclass or NULL if not a valid cast
      
  REVISIONS
    11/30/06 - J. P. - Modified to use ClassID class
    
  AUTHOR
    J. P. Abelanet - 12/08/05
*******************************************************************************/
{
  if (entity && entity->Is_InClassID (Entity::Static_ClassID ()))
    return (const Entity*)entity;
  return NULL;
} // Entity::Downcast


Entity* Entity::Downcast (Entity* entity)
/*******************************************************************************
  ABSTRACT
    This performs safe downcasting of subclasses of Entity.

  PARAMETERS
    entity  IN
      Entity to use
    
    returns:
      Entity cast to this subclass or NULL if not a valid cast
      
  REVISIONS
    11/30/06 - J. P. - Modified to use ClassID class
    
  AUTHOR
    J. P. Abelanet - 12/08/05
*******************************************************************************/
{
  if (entity && entity->Is_InClassID (Entity::Static_ClassID ()))
    return (Entity*)entity;
  return NULL;
} // Entity::Downcast


Database *Entity::Inquire_Database () const
/*******************************************************************************
  ABSTRACT
    This routine determines the Database that an Entity is attached to,
    if one exists.  An Entity can be attached to only one Database at
    a time.
  
  PARAMETERS
    returns:
      Database that Entity is attached to or NULL if it is not attached to one
  
  NOTES
    This only acknowledges Entities that are directly attached to a Database.
    Attributes, for example, are attached through their AttributeRegistries.
    
  REVISIONS
    04/28/01 - J. P. - Changed Entity from void* to void
    02/26/03 - J. P. - Modified to correctly handle unshared attributes
    10/14/04 - J. P. - Modified to use classes
    01/18/06 - J. P. - Modified to call Entity::Is_InClassID()
    02/27/06 - J. P. - Modified to return Database vs Error
    
  AUTHOR
    J. P. Abelanet - 03/23/01
*******************************************************************************/
{
  DatabaseReference	*databaseReference;

  /* If this is an (unshared) Attribute, it cannot possibly be attached to
     a Database - in fact, the subsequent routines will incorrectly conclude
     that this might be attached to a Database if its unique owning entity
     is attached to a Database */
  if (Is_InClassID (Attribute::Static_ClassID ()))
    return NULL;
  
  /* Find the Attribute */
  databaseReference = static_cast<DatabaseReference*> (Attribute::Entity_InquireByClass_FirstAttribute (this,
                  DatabaseReference::Static_ClassID ()));
  
  /* Find the Database pointer */
  if (databaseReference)
    return databaseReference->Inquire_Database ();

  return NULL;
} /* Entity::Inquire_Database */


Error Entity::Detach_Database () const
/*******************************************************************************
  ABSTRACT
    This routine detaches an entity from the database containing it.
    An error will result if the entity is not attached to a database.
  
  PARAMETERS
    returns:
      Error_invalid_entity	- Entity type not valid
      Error_database_missing	- Entity not attached to a Database
  
  REVISIONS
    03/25/97 - J. P. - Modified to call NMBDB_FreeDB
    05/03/98 - J. P. - Added support for groups
    01/24/99 - J. P. - Split BSp into CrvBSp and SurfBSp
    07/31/99 - J. P. - Restructured slightly to not call NMBDB_FreeDB
    11/21/00 - J. P. - Modified to call Array::Free
    01/31/01 - J. P. - Modified to treat database as an entity and to return
                       an error code
    03/23/01 - J. P. - Modified to use backpointers
    04/28/01 - J. P. - Changed Entity from void* to void
    02/26/03 - J. P. - Modified to correctly handle unshared attributes
    06/10/03 - J. P. - Modified to use linked lists of entities (vs arrays) and
                       to use templates
    10/14/04 - J. P. - Modified to use classes
    01/18/06 - J. P. - Modified to call Entity::Is_InClassID()
    11/30/06 - J. P. - Modified to use ClassID class
    
  AUTHOR
    J. P. Abelanet - 01/31/97
*******************************************************************************/
{
  DatabaseReference	*databaseReference;
  Database		*database = NULL;
  
  /* If this is an (unshared) attribute, it cannot possibly be attached to
     a database - in fact, the subsequent routines will incorrectly conclude
     that this might be attached to a database if its unique owning entity
     is attached to a database */
  if (Is_InClassID (Attribute::Static_ClassID ()))
    return Error_no_errors;
  
  /* Is this Entity already attached to a Database?
     If not, this is an error */
  databaseReference = static_cast<DatabaseReference*>(Attribute::Entity_InquireByClass_FirstAttribute (this,
                  DatabaseReference::Static_ClassID ()));
  if (databaseReference)
    database = databaseReference->Inquire_Database ();
  if (database == NULL) {
    fprintf (ErrorLogger::GetErrorFile (),
    	"INTERNAL ERROR:  Entity::Detach_Database:  Entity not attached to a Database!\n");
    return Error_database_missing;
  }

  // Find the owning Attribute group
  AttsInClassID	*attsInClassID;
  
  if ((attsInClassID = database->Inquire_AttsInClassID (Inquire_ClassID ())) == NULL) {
    fprintf (ErrorLogger::GetErrorFile (),
    	"INTERNAL ERROR:  Entity::Detach_Database:  Database is corrupt!\n");
    return Error_invalid_entity;
  }

  /* Remove this Attribute from the list of Attributes attached to Entities
     of the same ClassID in the same Database */
  attsInClassID->Remove_DatabaseReference (databaseReference);
    
  /* Delete the database attribute */
  delete databaseReference;
  return Error_no_errors;
} /* Entity::Detach_Database */


Error Entity::Delete ()
/*******************************************************************************
  ABSTRACT
    This permanently and irreversibly deletes an Entity.
    In the process of deleting the Entity, the following will occur:
    - The Entity will be detached from its optional owning Database
    - The Entity will be removed from any Group entities to which it belongs
    - SharedAttribute instances will be detached from the Entity
    - UnsharedAttribute instances will be deleted via their destructors
  
  PARAMETERS
    returns:
      Error_invalid_entity	- entity type not valid
      Error_invalid_entity	- entity not in database
      Error_still_referenced	- entity still referenced by another entity
  
  NOTES
    - If an Entity has transient parents that, after deletion of the Entity,
      have no dependent children, then they will be deleted as well
      - Deleting a CurvedFace will delete the support Surface if it is not
        bound to anyone
    - Registers a DELETE event on the Entity
    - Currently can be any sub-class of Entity other than:
      - Subclasses of Topology other than Model
      - BSplineCurve in parameter space
      - Transform
      - Database
  
  BUGS
    A more general Entity:: method that indicates that the Entity is about to
    be deleted should be called by this method and Edge::Op_Kill(), etc -
    maybe all the Attribute related calls should be handled in there.
    This would include the check for whether an Entity is strongly bound
    (Inquire_Deletable()).
  
  REVISIONS
    03/23/01 - J. P. - Modified to use backpointers
    04/28/01 - J. P. - Changed Entity from void* to void
    07/28/01 - J. P. - Modified to detach entity from groups and to check
                       for remaining associativity
    08/02/01 - J. P. - Added support for circles, revolutions, and extrusions
    10/02/02 - J. P. - Modified to call Entity::Delete
    02/26/03 - J. P. - Modified to handle attributes and attribute templates
    02/28/03 - J. P. - Modified to pass Entity::Static_ClassID () to StronglyBindingChild::Entity_Inquire_Children
    03/10/03 - J. P. - Added explicit casts
    03/26/03 - J. P. - Swapped parameters to Group::Attach/Detach_Entity
    05/06/04 - J. P. - Adding explicit casts in preparation for conversion to
                       C++ classes
    11/11/04 - J. P. - Modified to call Internal_Delete method
    03/16/05 - J. P. - Modified to set DeletePending_ flag
    04/28/05 - J. P. - Modified to use EntityList for all lists of entities
    02/27/06 - J. P. - Simplified greatly to use more Attribute change methods
    
  AUTHOR
    J. P. Abelanet - 02/11/01
*******************************************************************************/
{
  Error	error = Error_no_errors;
  
  /* If this is an Attribute, bypass all additional processing since
     Attributes cannot have Attributes */
  if (Is_InClassID (Attribute::Static_ClassID ())) {
    delete this;
    return Error_no_errors;
  }
  
  // Does this Entity have any strongly binding children?
  //   If so, it is an error to be in here!
  Entity	*DependentChild;
  if (!Inquire_Deletable (&DependentChild)) {
    fprintf (ErrorLogger::GetErrorFile (),
        "ERROR:  Entity::Delete:  entity %p (%s) still referenced by another entity %p (%s)!\n",
	this, Inquire_Description (),
	DependentChild, DependentChild->Inquire_Description ());
    return Error_still_referenced;
  }
  
  /* Delete the Attributes */
  if ((error = AttributeChangeNotifier::Entity_Modify_Attributes (this,
                   NULL, AttributeChangeNotifier::EntOp_Delete)) != Error_no_errors)
    return (error);
  
  /* Destroy the Entity */
  return Internal_Delete ();
} /* Entity::Delete */


Error Entity::Delete_Multiple (EntityList<Entity> *const entityList)
/*******************************************************************************
  ABSTRACT
    This routine deletes multiple Entities.  This takes into
    account dependencies between Entities and deletes them in the proper order.
    If an Entity in this list has dependent children, they must also appear in
    the list or an error will occur.
  
  PARAMETERS
    entityList  INOUT
      entityList to destroy - on return will only contain Entities that could
      not be deleted

    returns:
      Error_invalid_entity	- entity type not valid
      Error_invalid_entity	- entity not in database
      Error_still_referenced	- entity still referenced by another entity

  NOTES
    - See Entity::Delete() for additional information
  
  REVISIONS
    11/16/04 - J. P. - Modified to use classes
    11/17/04 - J. P. - Modified to be static and to modify the incoming list
                       of entities to reflect deletions
    12/16/04 - J. P. - Modified to use EntityList<Entity> class
    03/22/05 - J. P. - Modified to use AssociativeEntityList class
    05/19/05 - J. P. - Modified to use EntityReferenceList template class
    11/21/06 - J. P. - Standardized methods on Entity container classes
    04/27/07 - J. P. - Added logic to prevent infinite loop
    
  AUTHOR
    J. P. Abelanet - 03/25/03
*******************************************************************************/
{
  EntityReferenceList<Entity>	localEntityList (*entityList);
  Entity		*DependentChild;
  Int32			i;
  Boolean		Skipped, Deleted;
  Error			localError, error = Error_no_errors;

  // Copy the list to a local list to use EntityReference objects
  entityList->Clear ();
  
  /* Try repeatedly to delete entities until (hopefully) none are left */
  do {
    Skipped = Deleted = FALSE;
    for (i = 0; i < localEntityList.Size (); i++)
      
      // Does this entity still exist?
      if (localEntityList[i])
    
	/* Can this entity be deleted? */
	if (localEntityList[i]->Inquire_Deletable (&DependentChild)) {
	  Deleted = TRUE;

	  // Delete this entity - it will be removed automatically from the list,
	  //   and other deleted entities will be removed as well
	  if ((localError = localEntityList[i]->Delete ()) != Error_no_errors)
            error = localError;
	} else
          Skipped = TRUE;
  } while (Skipped && Deleted);
  
  /* Are there any entities left? */
  if (Skipped) {
    error = Error_still_referenced;
    localEntityList.AppendTo_EntityList (entityList);
    for (i = 0; i < entityList->Size (); i++) {
      (*entityList)[i]->Inquire_Deletable (&DependentChild);
      fprintf (ErrorLogger::GetErrorFile (),
          "ERROR:  Entity::Delete_Multiple:  Entity (%s) still referenced by another Entity (%s)!\n",
	  (*entityList)[i]->Inquire_Description (),
	  DependentChild->Inquire_Description ());
    }
  }
    
  return (error);
} /* Entity::Delete_Multiple */


Boolean Entity::Inquire_Deletable (Entity **const DependentChild) const
/*******************************************************************************
  ABSTRACT
    This routine determines if an Entity can be deleted by looking for strongly
    binding children, if any exist.
  
  PARAMETERS
    DependentChild  OUT
      If non-NULL, this is a strongly binding child of the Entity that
      prevents it from being deleted
      
    returns:
      TRUE if Entity has no strongly binding children
  
  REVISIONS
    04/23/03 - J. P. - Modified to check for dependent instances
    11/16/04 - J. P. - Modified to use classes
    04/28/05 - J. P. - Modified to use EntityList for all lists of entities
    08/24/05 - J. P. - Modified to call Group::Inquire_Instance()
    04/18/06 - J. P. - Greatly simplified to call
                       StronglyBindingChild::Entity_Inquire_FirstChild()
    
  AUTHOR
    J. P. Abelanet - 02/11/01
*******************************************************************************/
{
  /* Find any strongly binding child */
  *DependentChild = StronglyBindingChild::Entity_Inquire_FirstChild (this,
  			Entity::Static_ClassID ());
  return (*DependentChild == NULL);
} /* Entity::Inquire_Deletable */


Attribute *Entity::Inquire_FirstAttribute (const AttributeRegistry *attributeRegistry) const
/*******************************************************************************
  ABSTRACT
    This routine finds the first Attribute in a specified AttributeRegistry
    attached to a given Entity.
  
  PARAMETERS
    attributeRegistry  IN
      AttributeRegistry containing the Attribute to find -
      if NULL, all Attributes will be considered
    
    returns:
      Attribute in the specified AttributeRegistry attached to the given Entity,
      if one exists
      
  NOTES
    - "this" Entity cannot be an UnsharedAttribute
    
  REVISIONS
    04/28/01 - J. P. - Changed Entity from void* to void
    02/25/03 - J. P. - Modified to use attribute templates
    11/16/04 - J. P. - Modified to use classes
    02/27/06 - J. P. - Modified to return Attribute vs Error
    
  AUTHOR
    J. P. Abelanet - 01/25/01
*******************************************************************************/
{
  return Attribute::Entity_Inquire_FirstAttribute (this, attributeRegistry);
} /* Entity::Inquire_FirstAttribute */


void Entity::Inquire_Attributes (const AttributeRegistry *attributeRegistry,
           EntityList<Attribute> *const attributes,
	   Int32 *const numAttributes) const
/*******************************************************************************
  ABSTRACT
    This routine finds all the Attributes in a specified AttributeRegistry
    that are attached to a given Entity, or counts the number of Attributes
    in a specified AttributeRegistry that are attached to a given Entity.
  
  PARAMETERS
    attributeRegistry  IN
      AttributeRegistry containing the Attributes to find -
      if NULL, all Attributes will be considered
    
    attributes  INOUT
      List of Attributes attached to the Entity
  
    numAttributes  OUT
      Optional parameter that defaults to NULL -
      if non-NULL, attributes will not be computed, and this will be the
      # of Attributes in the specified AttributeRegistry attached to the given
      Entity - will be at most 1 for a specific AttributeRegistry unless the
      AttributeRegistry supports multiple instantiation
      
  NOTES
    - For a given AttributeRegistry, an Entity can only have more than one
      Attribute in that AttributeRegistry if it supports multiple instantiation.
    - "this" Entity cannot be an UnsharedAttribute
    
  REVISIONS
    04/28/01 - J. P. - Changed Entity from void* to void
    02/25/03 - J. P. - Modified to use attribute templates
    11/16/04 - J. P. - Modified to use classes
    11/16/06 - J. P. - Modified to return EntityList
    11/21/06 - J. P. - Removed Error return value from simple Inquire methods
    
  AUTHOR
    J. P. Abelanet - 01/25/01
*******************************************************************************/
{
  Attribute::Entity_Inquire_Attributes (this, attributeRegistry,
              attributes, numAttributes);
} /* Entity::Inquire_Attributes */


void Entity::Print_Attributes (const AttributeRegistry *attributeRegistry,
           FILE *strm) const
/*******************************************************************************
  ABSTRACT
    This routine prints all the Attributes in a specified AttributeRegistry
    that are attached to a given Entity.
  
  PARAMETERS
    attributeRegistry  IN
      AttributeRegistry containing the Attributes to find -
      if NULL, all Attributes will be considered
    
    strm  INOUT
      Stream to print diagnostics to
      
  AUTHOR
    J. P. Abelanet - 04/30/07
*******************************************************************************/
{
  EntityList<Attribute>	attributes;
  Int32			i;
  
  // Get a list of all the Attributes attached to this Entity
  Inquire_Attributes (attributeRegistry, &attributes);
  
  // Print them out
  if (attributeRegistry)
    fprintf (strm, "Entity %p (%s) - Attributes from AttributeRegistry %p\n",
	   this, Inquire_Description (), attributeRegistry);
  else
    fprintf (strm, "Entity %p (%s) - Attributes\n",
	   this, Inquire_Description ());
  for (i = 0; i < attributes.Size (); i++) {
    fprintf (strm, "\tAtt[%d] = %p (%s), AttributeRegistry = %p\n",
	   i, attributes[i], attributes[i]->Inquire_Description (),
	   attributes[i]->Inquire_AttributeRegistry ());
    fprintf (strm, "\t\tshared = %s, binding = %s\n",
	   attributes[i]->Inquire_Shared ()?"TRUE":"FALSE",
	   attributes[i]->Inquire_Binding ()?"TRUE":"FALSE");
  }
} // Entity::Print_Attributes


Boolean Entity::Inquire_Bound () const
/*******************************************************************************
  ABSTRACT
    Determines if an Entity is bound.
  
  PARAMETERS
    returns:
      TRUE if entity bound, and hence cannot be automatically deleted
  
  NOTES
    An Entity is normally considered bound if it has any subclasses of
    BindingAttributes attached to it.
    
  AUTHOR
    J. P. Abelanet - 03/31/05
*******************************************************************************/
{
  UnsharedAttribute	*LocalAttribute = Inquire_Attribute ();

  while (LocalAttribute)
    if (LocalAttribute->Inquire_Binding ())
      return TRUE;
    else
      LocalAttribute = LocalAttribute->Inquire_Attribute ();

  return FALSE;
} // Entity::Inquire_Bound


Error Entity::Copy_Multiple (const EntityList<Entity> *entityList,
         EntityList<Entity> *const EntityCopies, const Boolean shareSupportEntities)
/*******************************************************************************
  ABSTRACT
    This routine copies mutiple Entities simultaneuously,
    accounting for shared support Entities.
      
  PARAMETERS
    entityList  IN
      Entities to copy
      
    EntityCopies  OUT
      Copies of the original Entities, one-to-one with entityList
      
    shareSupportEntities  IN
      Optional parameter that defaults to FALSE -
      If TRUE, share the support Curves, Surfaces, etc -
      should almost always be FALSE
  
    returns:
      Error_insufficient_memory	- unable to allocate memory
      Error_general_unknown	- internal hashing error
  
  NOTES
    - Additional Entities may be created as needed
    - Registers the following events:
      - COPY event on each original and corresponding new Entity
  
  REVISIONS
    12/14/04 - J. P. - Modified UIDEN_xxx to return Error
    12/16/04 - J. P. - Modified to use EntityList<Entity> class
    08/17/05 - J. P. - Modified to use Hash class
    10/24/05 - J. P. - Modified to use EntityCopier class and to register COPY
                       events
    04/29/06 - J. P. - Now properly handles all Entities when
                       shareSupportEntities == TRUE
    
  AUTHOR
    J. P. Abelanet - 11/28/04
*******************************************************************************/
{
  return EntityCopier::Copy_Multiple (entityList, EntityCopies, shareSupportEntities);
} /* Entity::Copy_Multiple */


Entity *Entity::Copy (const Boolean shareSupportEntities) const
/*******************************************************************************
  ABSTRACT
    This routine copies an Entity.

  PARAMETERS
    shareSupportEntities  IN
      Optional parameter that defaults to FALSE -
      If TRUE, share the support Curves, Surfaces, etc -
      should almost always be FALSE
  
    returns:
      Copy of Entity -
      will be NULL if this routine fails

  NOTES
    - Additional Entities may be created as needed
    - Registers the following events:
      - COPY event on each original and corresponding new Entity
  
  BUGS
    - Does not place new Entities in the same Database as the originals
    - New Entities created on behalf of the desired Entity should NOT be
      placed into any Database, allowing them to be automatically deleted if
      the top-level Entity is deleted!!
    
  REVISIONS
    03/10/03 - J. P. - Added explicit casts
    05/06/04 - J. P. - Adding explicit casts in preparation for conversion to
                       C++ classes
    11/28/04 - J. P. - Modified to use classes and to return the new entity
    12/16/04 - J. P. - Modified to use EntityList<Entity> class
    10/24/05 - J. P. - Modified to use EntityCopier class and to register COPY
                       events
    
  AUTHOR
    J. P. Abelanet - 10/02/02
*******************************************************************************/
{
  return EntityCopier::Copy (this, shareSupportEntities);
} /* Entity::Copy */


Error Entity::Apply_Transform_Multiple (EntityList<Entity> *entityList,
         const TransformMatrix& Matrix)
/*******************************************************************************
  ABSTRACT
    This routine transforms mutiple Entities simultaneuously,
    accounting for shared support Entities.
      
  PARAMETERS
    entityList  IN
      List of Entities to transform
  
    Matrix  IN
      Transformation matrix to apply
  
    returns:
      Error_invalid_entity	- entity type not valid
  
  NOTES
    - Support Entities may be copied as needed
    - Registers the following events:
      - TRANSFORM event on each Entity
  
  BUGS
    If an Instance is to be transformed, it must currently be the only Entity
    in the list, since interactions with Instances are not properly handled!
    
  AUTHOR
    J. P. Abelanet - 05/01/06
*******************************************************************************/
{
  return EntityTransformer::Apply_Transform_Multiple (entityList, Matrix);
} /* Entity::Apply_Transform_Multiple */


Error Entity::Apply_Transform (const TransformMatrix& Matrix)
/*******************************************************************************
  ABSTRACT
    This routine transforms an Entity.
  
  PARAMETERS
    Matrix  IN
      Transformation matrix to apply
  
    returns:
      Error_invalid_entity	- entity type not valid

  NOTES
    - Support Entities may be copied as needed
    - Registers a TRANSFORM event on the Entity
  
  REVISIONS
    06/29/94 - J. P. - Implemented Stride
    10/03/99 - J. P. - Modified to call Memory::Free (vs free)
    03/10/03 - J. P. - Added explicit casts
    11/12/04 - J. P. - Modified to use classes
    08/24/05 - J. P. - Modified to use TransformMatrix class
    05/01/06 - J. P. - Modified to use EntityTransformer to handle Entity
                       interactions
      
  AUTHOR
    J. P. Abelanet - 10/02/02
*******************************************************************************/
{
  return EntityTransformer::Apply_Transform (this, Matrix);
} /* Entity::Apply_Transform */


Boolean Entity::Is_Major () const
/*******************************************************************************
  ABSTRACT
    This determines if an Entity is "major", meaning it can be a attached to
    a Database.
      
  PARAMETERS
    returns:
      TRUE if this is a major Entity
  
  NOTES
    - A major entity is any sub-class of Entity other than:
      - Attribute and its sub-classes
      - Topology subclasses other than CurvedModel and MeshModel
      - BSplineCurve in parameter space
      - Transform
      - Database
      - Cached Entities
    
  REVISIONS
    11/30/06 - J. P. - Made a virtual method
    
  AUTHOR
    J. P. Abelanet - 01/28/97
*******************************************************************************/
{
  return TRUE;
} /* Entity::Is_Major */


}	// namespace GE
